/* The GIMP -- an image manipulation program
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * RAW plugin version 0.0.1 
 * Copyright (C) 1999 Jaroslav Benkovsky, <benkovsk@pha.pvt.cz>
 *
 * RCS: $Header: /home/benkovsk/worx/RAW/RCS/RAW.c,v 1.2 1999/01/31 22:34:15 benkovsk Exp benkovsk $
 */

/*
  (gimp-display-new (car (file-raw2-load 0 "demo/colorcube.raw" "demo/colorcube.raw" "" "")))

or

  (gimp-display-new (car (file-raw2-load 1 "demo/nonheader256-256-4.raw" "demo/nonheader256-256-4.raw" "nonheader" "256:256:0:4:1")))

or

  (gimp-display-new (car (file-raw2-load-nonheader 1 "demo/nonheader256-256-4.raw" "demo/nonheader256-256-4.raw" 256 256 0 4 1)))

or  (this one will crash)

  (gimp-display-new (car (file-raw2-load 1 "demo/nonheader256-256-4.raw" "demo/nonheader256-256-4.raw" "nonheader" "")))
 */

/**********************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "RAW.h"

/**********************************************************/
#define SCALE_WIDTH 125

/**********************************************************/
/* Structs for the format dialog */
typedef struct
{
  gchar   format_name[1000];
  gchar   format_args[DEF_MAX_ARG_SIZE];
} RawSaveVals;

typedef struct
{
  gchar   format_name[1000];
  gchar   format_args[DEF_MAX_ARG_SIZE];
} RawLoadVals;

typedef struct
{
  gint  run;
} RawSaveInterface;

/**********************************************************/
/* Declare local functions */
static void
query (void);

static void
run (char    *name,
     int      nparams,
     GParam  *param,
     int     *nreturn_vals,
     GParam **return_vals);

static gint32
load_image   (char   *filename);


static gint
save_image   (char   *filename,
              gint32  image_ID,
              gint32  drawable_ID);

static gint
do_format_dialog (ContextType context);

static gint
custom_dialog (ContextType context, RawFormat *format);


static void
selection_callback  (GtkList *list);

static gint
listitem_press_callback (GtkWidget *widget, 
			 GdkEventButton *event, 
			 gpointer data);

static void
close_callback  (GtkWidget *widget,
		 gpointer   data);

static void
ok_callback     (GtkWidget *widget,
		 gpointer   data);

static void
entry_callback  (GtkWidget *widget,
		 gpointer   data);

static void
scale_update    (GtkAdjustment *adjustment,
		 gpointer      data);

static void
toggle_callback (GtkWidget *widget,
		 gpointer   data);



/**********************************************************/
GPlugInInfo PLUG_IN_INFO =
{
  NULL,    /* init_proc */
  NULL,    /* quit_proc */
  query,   /* query_proc */
  run,     /* run_proc */
};


static RawSaveVals save_vals = 
{
  "",
  ""
};

static RawLoadVals load_vals = 
{
  "",
  ""
};

static RawSaveInterface rawint =
{
  FALSE   /*  run  */
};


/**********************************************************/
/**********************************************************/
#ifdef RAW_WITH_GUILE
int 
main (int argc, char *argv[])
{
  gh_enter (argc, argv, gimp_main);
  return 0;       /* never reached */
}
#else
MAIN ();
#endif


/**********************************************************/
/* registers RAW plugin and calls fn's to load and 
 * register all subplugins
 */

static void
query ()
{
  static GParamDef load_args[] =
  {
    { PARAM_INT32, "run_mode", "Interactive, non-interactive" },
    { PARAM_STRING, "filename", "The name of the file to load" },
    { PARAM_STRING, "raw_filename", "The name entered" },
    { PARAM_STRING, "format", "The RAW format name" },
    { PARAM_STRING, "subargs", "Subformat arguments" },
  };
  static GParamDef load_return_vals[] =
  {
    { PARAM_IMAGE, "image", "Output image" },
  };
  static int nload_args = sizeof (load_args) / sizeof (load_args[0]);
  static int nload_return_vals = sizeof (load_return_vals) / sizeof (load_return_vals[0]);

  static GParamDef save_args[] =
  {
    { PARAM_INT32,    "run_mode", "Interactive, non-interactive" },
    { PARAM_IMAGE,    "image", "Input image" },
    { PARAM_DRAWABLE, "drawable", "Drawable to save" },
    { PARAM_STRING,   "filename", "The name of the file to save the image in" },
    { PARAM_STRING,   "raw_filename", "The name of the file to save the image in" },
    { PARAM_STRING,   "format", "The RAW format name" },
    { PARAM_STRING,   "subargs", "Subplugin arguments" },
  };
  static int nsave_args = sizeof (save_args) / sizeof (save_args[0]);

  gimp_install_procedure (PLUGIN_LOAD_NAME,
                          "Loads files of various raw formats",
                          "This plug-in loads files in \"RAW\" formats which "
			  "are specified in definition files.",  /* FIXME */
                          "Jaroslav Benkovsky",
                          "Jaroslav Benkovsky <benkovsk@pha.pvt.cz>",
                          "January 1999",
                          PLUGIN_LOAD_MENU,
                          NULL,
                          PROC_PLUG_IN,
                          nload_args, nload_return_vals,
                          load_args, load_return_vals);
  
  gimp_install_procedure (PLUGIN_SAVE_NAME,
                          "Saves files in various raw file formats",
                          "This plug-in saves files in \"RAW\" formats which "
			  "are specified in definition files.",  /* FIXME */
                          "Jaroslav Benkovsky",
                          "Jaroslav Benkovsky <benkovsk@pha.pvt.cz>",
                          "January 1999",
                          PLUGIN_SAVE_MENU,
                          "*",
                          PROC_PLUG_IN,
                          nsave_args, 0,
                          save_args, NULL);

  /* FIXME: the magic is bogus */
  gimp_register_magic_load_handler (PLUGIN_LOAD_NAME, "raw", PLUGIN_LOAD_MENU,
				    "0,string,RAW");

  gimp_register_save_handler (PLUGIN_SAVE_NAME, "raw", PLUGIN_SAVE_MENU);

  rawformat_load_all ();
  subplugin_install_all ();
}

/**********************************************************/
static void
run (char    *name,
     int      nparams,
     GParam  *param,
     int     *nreturn_vals,
     GParam **return_vals)
{
  static GParam  values[2];
  GRunModeType   run_mode;
  gint32         image_ID;
  GStatusType    status = STATUS_EXECUTION_ERROR;
  gboolean       is_subplugin;


  run_mode = param[0].data.d_int32;

  *nreturn_vals = 1;
  *return_vals = values;

  values[0].type = PARAM_STATUS;
  values[0].data.d_status = STATUS_CALLING_ERROR;

  rawformat_load_all ();

  /************** LOAD *******************/
  if (strncmp (name, PLUGIN_LOAD_NAME, strlen (PLUGIN_LOAD_NAME)) == 0) {
    is_subplugin = (strlen (name) != strlen (PLUGIN_LOAD_NAME));

    if (is_subplugin)
      selected_format = rawformat_find (name + strlen (PLUGIN_LOAD_NAME) + 1);
    else
      selected_format = rawformat_find (param[3].data.d_string);


    switch (run_mode) {
    case RUN_INTERACTIVE:
      gimp_get_data (PLUGIN_LOAD_NAME, &load_vals);

      if (!is_subplugin || !selected_format) {
	if (!do_format_dialog (CTX_LOAD)) break;
      }
      if (!selected_format) break;

      if (!subplugin_get_args_from_pdb (selected_format, CTX_LOAD)) return;
	
      if (!custom_dialog (CTX_LOAD, selected_format)) break;

      status = STATUS_SUCCESS;
      break;  /* case RUN_INTERACTIVE */

    case RUN_NONINTERACTIVE:
      if (!selected_format) return;

      if (is_subplugin) {
	if (!subplugin_get_args (selected_format, CTX_LOAD, 
				 &param[3], nparams-3)) return;
      }
      else {
	if (nparams != 5) return;      
	else if (!subplugin_get_args_from_string (selected_format, CTX_LOAD, 
						  param[4].data.d_string)) return;
      }

      status = STATUS_SUCCESS;
      break;  /* case RUN_NONINTERACTIVE */

    case RUN_WITH_LAST_VALS:
      gimp_get_data (PLUGIN_LOAD_NAME, &load_vals);

      if (!is_subplugin || !selected_format) {
	rawformat_find (load_vals.format_name);
      }
      if (!selected_format) return;

      if (!subplugin_get_args_from_pdb (selected_format, CTX_LOAD)) return;
	
      status = STATUS_SUCCESS;
      break;  /* case RUN_INTERACTIVE */

    default:
      return;
    }

    if (status != STATUS_SUCCESS) {
      values[0].data.d_status = status;
      return;
    }

    image_ID = load_image (param[1].data.d_string);

    if (image_ID != -1) {
      gimp_set_data (PLUGIN_LOAD_NAME, &load_vals, sizeof (RawLoadVals));
      subplugin_set_args_to_pdb (selected_format, CTX_LOAD);

      *nreturn_vals = 2;
      values[0].data.d_status = STATUS_SUCCESS;
      values[1].type = PARAM_IMAGE;
      values[1].data.d_image = image_ID;
    }
    else values[0].data.d_status = STATUS_EXECUTION_ERROR;
  }  /* load */
  /************** SAVE *******************/
  else if (strncmp (name, PLUGIN_SAVE_NAME, strlen (PLUGIN_SAVE_NAME)) == 0) {
    is_subplugin = (strlen (name) != strlen (PLUGIN_SAVE_NAME));

    if (is_subplugin)
      selected_format = rawformat_find (name + strlen (PLUGIN_SAVE_NAME) + 1);
    else
      selected_format = rawformat_find (param[5].data.d_string);


    switch (run_mode) {
    case RUN_INTERACTIVE:
      gimp_get_data (PLUGIN_SAVE_NAME, &save_vals);

      if (!is_subplugin || !selected_format) {
	if (!do_format_dialog (CTX_SAVE))
	  break;
      }
      if (!selected_format) break;

      if (!subplugin_get_args_from_pdb (selected_format, CTX_SAVE)) return;

      if (!custom_dialog (CTX_SAVE, selected_format)) break;
      status = STATUS_SUCCESS;
      break;  /* case RUN_INTERACTIVE */
	
    case RUN_NONINTERACTIVE:
      if (!selected_format) return;

      if (is_subplugin) {
	if (!subplugin_get_args (selected_format, CTX_SAVE, 
				 &param[5], nparams-5)) return;
      }
      else {
	if (nparams != 7) return;
	else 
	  if (!subplugin_get_args_from_string (selected_format, CTX_SAVE, 
					       param[6].data.d_string)) return;
      }

      status = STATUS_SUCCESS;
      break;  /* case RUN_NONINTERACTIVE */
	  
    case RUN_WITH_LAST_VALS:
      gimp_get_data (PLUGIN_SAVE_NAME, &save_vals);

      if (!is_subplugin || !selected_format) {
	rawformat_find (save_vals.format_name);
      }
      if (!selected_format) return;

      if (!subplugin_get_args_from_pdb (selected_format, CTX_SAVE)) return;
	
      status = STATUS_SUCCESS;
      break;  /* case RUN_INTERACTIVE */

    default:
      return;
    }

    if (status != STATUS_SUCCESS) {
      values[0].data.d_status = status;
      return;
    }

    *nreturn_vals = 1;
    if (save_image (param[3].data.d_string,
		    param[1].data.d_int32,
		    param[2].data.d_int32))
      {
	gimp_set_data (PLUGIN_SAVE_NAME, &save_vals, sizeof (RawSaveVals));
	subplugin_set_args_to_pdb (selected_format, CTX_SAVE);

	values[0].data.d_status = STATUS_SUCCESS;
      }
    else {
      values[0].data.d_status = STATUS_EXECUTION_ERROR;
    }
  }
  else
    g_assert_not_reached ();
}

/**********************************************************/
static gint32
load_image (char *filename)
{
  gint32       image_ID = -1;

  char         *name;

  HeaderFn     header_load;
  PixelFn      pixel_load;
  DrawableFn   drawable_load;
  ImageFn      image_load;

  gint         rc = FALSE;
  FILE         *src = NULL;


  /* put up a progress bar */
  name = (char*)g_malloc (strlen (filename) + 12);
  sprintf (name, "Loading %s:", filename);
  gimp_progress_init (name);
  g_free (name);

  src = fopen (filename, "r");
  if (!src) goto cleanup;

  setitem_process_all (selected_format, CTX_LOAD, PHASE_BEFHDR);

  header_find_function (CTX_LOAD, selected_format->header_fn, &header_load);

  if (!header_load (src, selected_format)) goto cleanup;

  if (!variable_get_num ("width", &width)) goto cleanup;
  if (!variable_get_num ("height", &height)) goto cleanup;
  if (!variable_get_num ("type", &image_type)) image_type = RGB;
  if (!variable_get_num ("bpp", &bpp)) bpp = 3; /* FIXME: should not be here */
  if (!variable_get_num ("alpha", &alpha)) alpha = 0;

  /* create the new image */
  image_ID = gimp_image_new (width, height, image_type);
  gimp_image_set_filename (image_ID, filename);

  data_find_function (CTX_LOAD, 
		      selected_format->pixel_fn,
		      selected_format->drawable_fn, 
		      selected_format->image_fn,
		      &pixel_load, 
		      &drawable_load, 
		      &image_load);

  /* output */
  rc = image_load (src, drawable_load, pixel_load, image_ID, NULL);

cleanup:
  if (src) fclose (src);
  return image_ID;
}


/**********************************************************/
static gint
save_image (char   *filename,
            gint32  image_ID,
            gint32  drawable_ID)
{
  GDrawable    *drawable;
  
  HeaderFn     header_save;
  PixelFn      pixel_save;
  DrawableFn   drawable_save;
  ImageFn      image_save;

  gint         rc = FALSE;

  FILE         *tgt = NULL;
  gchar        *name;


  /* put up a progress bar */
  name = (gchar*)g_malloc (strlen (filename) + 12);
  sprintf (name, "Saving %s:", filename);
  gimp_progress_init (name);
  g_free (name);

  /* get some basic stats about the image */
  switch (gimp_drawable_type (drawable_ID)) {
  case RGBA_IMAGE:     alpha = 1; color = 1; indexed = 0; break;
  case INDEXEDA_IMAGE: alpha = 1; color = 1; indexed = 1; break;
  case GRAYA_IMAGE:    alpha = 1; color = 0; indexed = 0; break;
  case RGB_IMAGE:      alpha = 0; color = 1; indexed = 0; break;
  case INDEXED_IMAGE:  alpha = 0; color = 1; indexed = 1; break;
  case GRAY_IMAGE:     alpha = 0; color = 0; indexed = 0; break;
  default: return FALSE;
  }

  drawable = gimp_drawable_get (drawable_ID);
  width    = drawable->width;
  height   = drawable->height;
  bpp      = drawable->bpp;


  variables_init ();
  variable_num_add ("width", width);
  variable_num_add ("height", height);
  variable_num_add ("bpp", bpp);
  variable_num_add ("alpha", alpha);
  variable_num_add ("color", color);
  variable_num_add ("indexed", indexed);
  variable_num_add ("type", gimp_image_base_type (image_ID));
  variable_num_add ("layer_type", gimp_drawable_type (drawable_ID));

  variable_text_add ("file", filename, -1);

#if 0
  variable_num_add ("layercnt",...);
  variable_text_add ("frmfile", ...);
  variable_text_add ("frmname", ...);
#endif


  setitem_process_all (selected_format, CTX_SAVE, PHASE_BEFHDR);

  tgt = fopen (filename, "w");


  header_find_function (CTX_SAVE, selected_format->header_fn, &header_save);
  if (!header_save (tgt, selected_format)) goto cleanup;

  data_find_function (CTX_SAVE, 
		      selected_format->pixel_fn, 
		      selected_format->drawable_fn, 
		      selected_format->image_fn,
		      &pixel_save, 
		      &drawable_save, 
		      &image_save);
  
  /* output */
  rc = image_save (tgt, drawable_save, pixel_save, image_ID, drawable);
  
cleanup:
  if (tgt) fclose (tgt);
  /* clean up resources */  
  gimp_drawable_detach (drawable);
  return rc;
}

/**********************************************************/
static gint
do_format_dialog (ContextType context)
{
  GtkWidget  *dlg;
  GtkWidget  *button;
  GtkWidget  *frame;
  GtkWidget  *scrolled_win;
  GtkWidget  *list;
  GtkWidget  *list_item;
  gchar      **argv;
  gint       argc;

  argc = 1;
  argv = g_new (gchar *, 1);
  argv[0] = g_strdup (context==CTX_SAVE?"save":"load");

  rawint.run = FALSE;

  gtk_init (&argc, &argv);
  gtk_rc_parse (gimp_gtkrc ());

  dlg = gtk_dialog_new ();
  gtk_window_set_title (GTK_WINDOW (dlg), context==CTX_SAVE?"Save as RAW":"Load as RAW");
  gtk_window_position (GTK_WINDOW (dlg), GTK_WIN_POS_MOUSE);
  gtk_signal_connect (GTK_OBJECT (dlg), "destroy",
		      (GtkSignalFunc) close_callback,
		      NULL);

  /*  Action area  */
  button = gtk_button_new_with_label ("OK");
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
                      (GtkSignalFunc) ok_callback,
                      dlg);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->action_area), button, TRUE, TRUE, 0);
  gtk_widget_grab_default (button);
  gtk_widget_show (button);

  button = gtk_button_new_with_label ("Cancel");
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			     (GtkSignalFunc) gtk_widget_destroy,
			     GTK_OBJECT (dlg));
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->action_area), button, TRUE, TRUE, 0);
  gtk_widget_show (button);

  /*  parameter settings  */
  frame = gtk_frame_new ("Select RAW Format");
  gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_ETCHED_IN);
  gtk_container_border_width (GTK_CONTAINER (frame), 10);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->vbox), frame, TRUE, TRUE, 0);
  gtk_widget_show (frame);


  scrolled_win = gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_win),
                                  GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_container_add (GTK_CONTAINER (frame), scrolled_win);
  gtk_widget_set_usize (scrolled_win, 200, 200);

  list = gtk_list_new ();

  gtk_signal_connect (GTK_OBJECT (list), "selection_changed",
		      (GtkSignalFunc)selection_callback,
		      NULL);

  gtk_list_set_selection_mode (GTK_LIST (list), GTK_SELECTION_BROWSE);
  gtk_container_add (GTK_CONTAINER (scrolled_win), list);

  gtk_widget_show (list);
  gtk_widget_show (scrolled_win);

  { 
    GList  *flist = rawformat_list->next;
    RawFormat  *format;

    while (flist) {
      format = (RawFormat*)flist->data;
      flist = flist->next;

      list_item = gtk_list_item_new_with_label (format->name);

      gtk_signal_connect (GTK_OBJECT(list_item), "button_press_event",
			  (GtkSignalFunc)listitem_press_callback,
			  GTK_OBJECT (dlg));

      gtk_container_add (GTK_CONTAINER (list), list_item);
      gtk_object_set_user_data (GTK_OBJECT(list_item), format);
      gtk_widget_show (list_item);
    }
  }


  gtk_widget_show (dlg);

  gtk_main ();
  gdk_flush ();

  return rawint.run;
}

/**********************************************************/
static gint
custom_dialog (ContextType context, RawFormat *format)
{
  GtkWidget  *dlg;
  GtkWidget  *label;
  GtkWidget  *entry;
  GtkWidget  *toggle;
  GtkWidget  *button;
  GtkWidget  *scale;
  GtkObject  *scale_data;
  GtkWidget  *frame;
  GtkWidget  *table;
  GList      *alist;
  gchar      **argv;
  gint       argc;

  gchar      *name;
  gchar      *text;


  if ((!format->ask_cnt_save && context==CTX_SAVE)
      || (!format->ask_cnt_load && context==CTX_LOAD)) return 1;

  argc = 1;
  argv = g_new (gchar *, 1);
  argv[0] = g_strdup (context==CTX_SAVE?"save":"load"); /* FIXME: only once */

  rawint.run = FALSE;

  gtk_init (&argc, &argv);
  gtk_rc_parse(gimp_gtkrc());

  dlg = gtk_dialog_new ();
  gtk_window_set_title (GTK_WINDOW (dlg), "Format Parameters");
  gtk_window_position (GTK_WINDOW (dlg), GTK_WIN_POS_MOUSE);
  gtk_signal_connect (GTK_OBJECT (dlg), "destroy",
		      (GtkSignalFunc) close_callback,
		      NULL);

  /*  Action area  */
  button = gtk_button_new_with_label ("OK");
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
                      (GtkSignalFunc) ok_callback,
                      dlg);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->action_area), button, TRUE, TRUE, 0);
  gtk_widget_grab_default (button);
  gtk_widget_show (button);

  button = gtk_button_new_with_label ("Cancel");
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			     (GtkSignalFunc) gtk_widget_destroy,
			     GTK_OBJECT (dlg));
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->action_area), button, TRUE, TRUE, 0);
  gtk_widget_show (button);


  /* Parameter Settings */
  name = (gchar*)g_malloc (strlen (format->name) + 9);
  strcpy (name, "Format: ");
  strcat (name, format->name);
  label = gtk_label_new (name);
  gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->vbox), label, TRUE, TRUE, 0);
  gtk_widget_show (label);
  g_free (name);

  frame = gtk_frame_new ("Parameter Settings");
  gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_ETCHED_IN);
  gtk_container_border_width (GTK_CONTAINER (frame), 10);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->vbox), frame, TRUE, TRUE, 0);

  table = gtk_table_new (g_list_length (format->ask_list), 2, FALSE);
  gtk_container_border_width (GTK_CONTAINER (table), 10);
  gtk_container_add (GTK_CONTAINER (frame), table);


  {
    gint    line;
    GtkWidget *widget;

    alist = format->ask_list->next;
    line = 0;
    while (alist) {
      AskItem  *ask = (AskItem*)alist->data;

      alist = alist->next;
      widget = NULL;
      if (ask->context != context && ask->context != CTX_BOTH) continue;

      /* argument label */
      label = gtk_label_new (ask->label->text);
      gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
      gtk_table_attach (GTK_TABLE (table), label, 0, 1, line+1, line+2, 
			GTK_FILL | GTK_EXPAND, GTK_FILL, 5, 0);
      gtk_widget_show (label);

      /* argument value */
      switch (ask->type) {
      case ASK_ENTRY:
	entry = gtk_entry_new ();
	text = expression_value_to_string (ask->value);
	gtk_entry_set_text (GTK_ENTRY (entry), text);
	g_free (text);
     	gtk_signal_connect (GTK_OBJECT (entry), "changed",
			    (GtkSignalFunc) entry_callback,
			    ask);
	widget = entry;
	break;

      case ASK_SCALE: 
	scale_data = gtk_adjustment_new (ask->value->fnum, 0.0, 1.0, 0.01, 0.01, 0.0);
	scale = gtk_hscale_new (GTK_ADJUSTMENT (scale_data));
	gtk_widget_set_usize (scale, SCALE_WIDTH, 0);
	gtk_scale_set_value_pos (GTK_SCALE (scale), GTK_POS_TOP);
	gtk_scale_set_digits (GTK_SCALE (scale), 2);
	gtk_range_set_update_policy (GTK_RANGE (scale), GTK_UPDATE_DELAYED);
	gtk_signal_connect (GTK_OBJECT (scale_data), "value_changed",
			    (GtkSignalFunc) scale_update,
			    ask);
	widget = scale;
	break;

      case ASK_TOGGLE:
	toggle = gtk_check_button_new ();
	gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (toggle), 
				     ask->value->num!=0);
	gtk_signal_connect (GTK_OBJECT (toggle), "clicked",
			    (GtkSignalFunc) toggle_callback,
			    ask);
	widget = toggle;
	break;

      default:
	g_assert_not_reached ();
      }

      if (widget) {
	gtk_table_attach (GTK_TABLE (table), widget, 1, 2, line+1, line+2, 
			  GTK_FILL | GTK_EXPAND, GTK_FILL, 5, 0);
	gtk_widget_show (widget);
      }

      line++;
    }
  }

  gtk_widget_show (table);
  gtk_widget_show (frame);
  gtk_widget_show (dlg);


  gtk_main ();
  gdk_flush ();

  subplugin_set_vars (format, context);

  return rawint.run;
}

/**********************************************************/
/*  Interface functions  */
static void
selection_callback (GtkList *list)
{
  GtkListItem *item;
  GList       *selected;

  selected = list->selection;
  if (selected==NULL) return;

  item = GTK_LIST_ITEM (selected->data);
  if (gtk_object_get_user_data (GTK_OBJECT (item))) {
    selected_format = gtk_object_get_user_data (GTK_OBJECT (item));
  }
  else {
    selected_format = (RawFormat*)rawformat_list->next->data;
  }
}

/**********************************************************/
static gint
listitem_press_callback (GtkWidget *widget, 
			 GdkEventButton *event, 
			 gpointer data)
{
  switch (event->type) {
  case GDK_2BUTTON_PRESS:
    ok_callback (NULL, data);
    break;

  default:
    break;
  }

  return FALSE;
}

/**********************************************************/
static void
close_callback (GtkWidget *widget,
		gpointer   data)
{
  gtk_main_quit ();
}

/**********************************************************/
static void
ok_callback (GtkWidget *widget,
	     gpointer   data)
{
  rawint.run = TRUE;
  gtk_widget_destroy (GTK_WIDGET (data));
}

/**********************************************************/
static void
entry_callback (GtkWidget *widget,
		gpointer   data)
{
  AskItem   *ask = (AskItem*) data;
  gchar     *text = gtk_entry_get_text (GTK_ENTRY (widget));


  /* FIXME: apply constraints here */
  expression_assign_value_from_string (ask->value, text);
}

/**********************************************************/
static void
toggle_callback (GtkWidget *widget,
		 gpointer   data)
{
  AskItem   *ask = (AskItem*) data;

  ask->value->num = GTK_TOGGLE_BUTTON (widget)->active;
}

/**********************************************************/
static void
scale_update (GtkAdjustment *adjustment,
	      gpointer      data)
{
  AskItem   *ask = (AskItem*) data;

  ask->value->fnum = adjustment->value;
}

/**********************************************************/
/* End of file RAW.c */
