/* The GIMP -- an image manipulation program
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * RAW plugin version 0.0.1
 * Copyright (C) 1999 Jaroslav Benkovsky, <benkovsk@pha.pvt.cz>
 *
 * RCS: $Header: /home/benkovsk/worx/RAW/RCS/data.c,v 1.2 1999/01/31 22:33:28 benkovsk Exp benkovsk $
 */

/**********************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "RAW.h"

/**********************************************************/
static void
pixel_save_simple (guchar *out_buf, guchar *in_buf, 
		   gint *olen, gint *ilen, 
		   OperType op);

static void
pixel_save_A2a (guchar *out_buf, guchar *in_buf, 
		gint *olen, gint *ilen,
		OperType op);


static void
pixel_load_A2a (guchar *out_buf, guchar *in_buf, 
		gint *olen, gint *ilen,
		OperType op);

static void
pixel_save_rgbacube (guchar *out_buf, guchar *in_buf, 
		     gint *olen, gint *ilen,
		     OperType op);

static void
pixel_load_rgbacube (guchar *out_buf, guchar *in_buf, 
		     gint *olen, gint *ilen,
		     OperType op);

static gint
drawable_save_lrud (FILE *tgt, PixelFn pixel_save, GDrawable *drawable);

static gint
drawable_load_lrud (FILE *src, PixelFn pixel_load, GDrawable *drawable);

static gint
image_save_bt (FILE *tgt, DrawableFn drawable_save, PixelFn pixel_save, gint32 image_ID, GDrawable *drawable);

static gint
image_save_tb (FILE *tgt, DrawableFn drawable_save, PixelFn pixel_save, gint32 image_ID, GDrawable *drawable);

static gint
image_save_single (FILE *tgt, DrawableFn drawable_save, PixelFn pixel_save, gint32 image_ID, GDrawable *drawable);

static gint
image_load_bt (FILE *src, DrawableFn drawable_load, PixelFn pixel_load, gint32 image_ID, GDrawable *drawable);

static gint
image_load_tb (FILE *src, DrawableFn drawable_load, PixelFn pixel_load, gint32 image_ID, GDrawable *drawable);

static gint
image_load_single (FILE *src, DrawableFn drawable_load, PixelFn pixel_load, gint32 image_ID, GDrawable *drawable);


/**********************************************************/
/* FIXME: should be dynamic/local ? */
int       width;
int       height;
int       image_type;
int       alpha;
int 	  ncolors=1;
int	  *indexno;
int 	  indexed;
int       bpp;
int       color;
int       alpha_threshold = 128;


/* jak ale do hlavicky dostat, co dana pixel funkce udelala s bpp apod? */

/**********************************************************/
/* simple: save/load pixel exactly same as it is in memory */
/* A2a: pixel in mem:true alpha, pixel in file:fake alpha (0 = transp) */
/* rgbacube: transform pixel channels to color cube before saving & v.v. */

PixelFnRec  pixel_fn_rec[] =
{
  /* name       save fn              load fn */
  { "simple",   pixel_save_simple,   pixel_save_simple },
  { "A2a",      pixel_save_A2a,      pixel_load_A2a },
  { "rgbacube", pixel_save_rgbacube, pixel_load_rgbacube },
};

gint  pixel_fn_rec_cnt = sizeof (pixel_fn_rec) / sizeof (PixelFnRec);


/**********************************************************/
/* lrud: save/load pixels in drawable left->right then up->down */

DrawableFnRec  drawable_fn_rec[] =
{
  /* name      save fn             load fn */
  { "lrud",    drawable_save_lrud, drawable_load_lrud }, 
};

gint  drawable_fn_rec_cnt = sizeof (drawable_fn_rec) / sizeof (DrawableFnRec);


/**********************************************************/
/* single: save/load active drawable only */
/* bt: save/load layers bottom->top */
/* tb: save/load layers top->bottom */

ImageFnRec  image_fn_rec[] =
{
  /* name      save fn             load fn */
  { "single",  image_save_single,  image_load_single }, 
  { "bt",      image_save_bt,      image_load_bt }, 
  { "tb",      image_save_tb,      image_load_tb }, 
};

gint  image_fn_rec_cnt = sizeof (image_fn_rec) / sizeof (ImageFnRec);


/**********************************************************/
/**********************************************************/
void
data_find_function (ContextType context,
		    gchar *pixel_name, gchar *drawable_name, gchar *image_name,
		    PixelFn *pixel_fn, DrawableFn *drawable_fn, ImageFn *image_fn)
{
  gint    i;
  gboolean  Save;


  g_assert (context == CTX_SAVE || context == CTX_LOAD);

  Save = (context == CTX_SAVE);

  if (pixel_fn) {
    *pixel_fn = Save ? pixel_fn_rec[0].save_fn : pixel_fn_rec[0].load_fn;

    if (pixel_name) {
      for (i=0; i<pixel_fn_rec_cnt; i++)
	if (!strcmp (pixel_name, pixel_fn_rec[i].name)) {
	  *pixel_fn = Save ? pixel_fn_rec[i].save_fn : pixel_fn_rec[i].load_fn;
	  break;
	}
    }
  }

  if (drawable_fn) {
    *drawable_fn = Save ? drawable_fn_rec[0].save_fn : drawable_fn_rec[0].load_fn;

    if (drawable_name) {
      for (i=0; i<drawable_fn_rec_cnt; i++)
	if (!strcmp (drawable_name, drawable_fn_rec[i].name)) {
	  *drawable_fn = Save ? drawable_fn_rec[i].save_fn : drawable_fn_rec[i].load_fn;
	  break;
	}
    }
  }

  if (image_fn) {
    *image_fn = Save ? image_fn_rec[0].save_fn : image_fn_rec[0].load_fn;

    if (image_name) {
      for (i=0; i<image_fn_rec_cnt; i++)
	if (!strcmp (image_name, image_fn_rec[i].name)) {
	  *image_fn = Save ? image_fn_rec[i].save_fn : image_fn_rec[i].load_fn;
	  break;
	}
    }
  }
}


/**********************************************************/
/**********************************************************/
static void
pixel_save_simple (guchar *out_buf, guchar *in_buf, 
		   gint *olen, gint *ilen, 
		   OperType op)
{
  if (op != OP_NORMAL) {
    *olen = *ilen = 0;
  }
  else {
    memcpy (out_buf, in_buf, bpp);
    *olen = *ilen = bpp;
  }
}

/**********************************************************/
/*
 * A2a :  true alpha -> fake alpha (0 = transparent)
 */

static void
pixel_save_A2a (guchar *out_buf, guchar *in_buf, 
		gint *olen, gint *ilen,
		OperType op)
{
  if (op != OP_NORMAL) {
    *olen = *ilen = 0;
    return;
  }

  *olen = *ilen = bpp;

  if (alpha) {
    if (in_buf[bpp-1] < alpha_threshold)
      memset (out_buf, 0, bpp-1);
    else {
      out_buf[0] = MIN (1 + (gint)in_buf[0], 0xFF);
      if (bpp>2) memcpy (out_buf+1, in_buf+1, bpp-2);
    }
    *olen = bpp-1;
  }
  else memcpy (out_buf, in_buf, bpp);
}

/**********************************************************/
static void
pixel_load_A2a (guchar *out_buf, guchar *in_buf, 
		gint *olen, gint *ilen,
		OperType op)
{
  if (op != OP_NORMAL) {
    *olen = *ilen = 0;
    return;
  }

  *olen = *ilen = bpp;
  if (alpha) {
    if (in_buf[0] == 0) {
      memset(out_buf, 0, bpp);
    }
    else {
      out_buf[0]=in_buf[0]-1;
      if (bpp>2) memcpy (out_buf+1, in_buf+1, bpp-2);
      out_buf[bpp-1] = 255;
    }
    *ilen = bpp-1;
  }
  else memcpy (out_buf, in_buf, bpp);
}

/**********************************************************/
/*
 * assumption: each out channel <0..8>
 *             4 out channels
 *             all drawables have the same bpp :-(
 *             in out buf, R channel is first and A channel is last
 */

static void
pixel_save_rgbacube (guchar *out_buf, guchar *in_buf, 
		     gint *olen, gint *ilen,
		     OperType op)
{
  static gint  len = 0, num_channels = 0;
  static gint  size[10], dlen;
  static gint cube;
  gint   i, num;
  guint32  buf;

  switch (op) {
  case OP_INIT:
    memset (size, 0, sizeof (size));
    if (!variable_get_num ("colorcube", &cube)) cube = 5551;

    while (cube) {
      num = cube % 10;
      cube = cube / 10;

      size[num_channels] = num;

      len += num;
      num_channels++;
    }
    dlen = (len+7) / 8;
    num_channels = MAX (num_channels, bpp);
    *olen = *ilen = 0;
    break;

  case OP_NORMAL:
    buf = 0;
    for (i=0; i<num_channels; i++) {
      buf <<= size[num_channels-i-1];
      buf |= ((guint)(i<bpp?in_buf[i]:255)) >> (8 - size[num_channels-i-1]);
    }

    for (i=0; i<dlen; i++) {
      num = buf % 256;
      buf /= 256;

      out_buf[dlen-i-1] = num;
    }
    *olen = dlen;
    *ilen = bpp;
    break;

  default:
    *olen = *ilen = 0;
  }
}

/**********************************************************/
static void
pixel_load_rgbacube (guchar *out_buf, guchar *in_buf, 
		     gint *olen, gint *ilen,
		     OperType op)
{
  static gint  len = 0, num_channels = 0;
  static gint  size[10], dlen;
  static gint cube;
  gint   i, j, num;
  guint32  buf;

  switch (op) {
  case OP_INIT:
    memset (size, 0, sizeof (size));
    if (!variable_get_num ("colorcube", &cube)) cube = 5551;

    while (cube) {
      num = cube % 10;
      cube = cube / 10;

      size[num_channels] = num;

      len += num;
      num_channels++;
    }
    dlen = (len+7) / 8;
    num_channels = MAX (num_channels, bpp);
    *olen = *ilen = 0;
    break;

  case OP_NORMAL:
    buf = 0;
    for (i=0; i<dlen; i++) {
      buf = buf*256 + in_buf[i];
    }

    for (i=0; i<num_channels; i++) {
      if (num_channels - i <= bpp) {
	num = (buf << (8 - size[i])) & 255;

	/* this is to get better color distribution */
	j=size[i];
	while (num && j && (j < 8)) {
	  num |= (num >> j);
	  j*=2;
	}

	out_buf[num_channels-i-1] = num;
      }
      buf >>= size[i];
    }
    *ilen = dlen;
    *olen = bpp;
    break;

  default:
    *olen = *ilen = 0;
  }
}

/**********************************************************/
#if 0
static void
pixel_save_xx ()
{
              /* get pixel data */
              key->r = *(data++);
	      key->g = color && !indexed ? *(data++) : key->r;
	      key->b = color && !indexed ? *(data++) : key->r;
	      a = alpha ? *(data++) : 255;
	      
	      if (a < threshold) 
		*(idata++) = 0;
	      else
		if (indexed)
		  *(idata++) = (key->r)+1;
		else {
		  indexno = g_hash_table_lookup(hash, key);
		  if (!indexno) {
		    indexno = g_new(int, 1);
		    *indexno = ncolors++;
		    g_hash_table_insert(hash, key, indexno);
		    key = g_new(rgbkey, 1);
		  }
		  *(idata++) = *indexno;
		}
}
#endif

/**********************************************************/
/*
 * lrud - left to right, then up to down
 */

static gint
drawable_save_lrud (FILE *tgt, PixelFn pixel_save, GDrawable *drawable)
{
  GPixelRgn      pixel_rgn;

  guchar         *in_buf = NULL;
  guchar         *in_buf_pos;
  guchar         *out_buf = NULL;
  guchar         *out_buf_pos;
  guchar         *out_buf_threshhold;

  gboolean       rc = FALSE;
  gint           i, j, k, olen, ilen;

  i = gimp_tile_height()*width*4;             /* g()*MAX(height,width)*4 ? */
  if ((out_buf = g_new(guchar, i)) == NULL)
    goto cleanup;

  out_buf_pos = out_buf;
  out_buf_threshhold = &out_buf[i-5];

  i = gimp_tile_height()*width*bpp;           /* g()*MAX(height,width)*bpp ? */
  if ((in_buf = g_new(guchar, i)) == NULL)
    goto cleanup;

  /* allocate a pixel region to work with */
  gimp_pixel_rgn_init (&pixel_rgn, drawable,
                       0, 0,
                       width, height,
                       TRUE, FALSE);

  pixel_save (out_buf_pos, NULL, &olen, &ilen, OP_INIT);
  out_buf_pos += olen;

  /* process each row of tiles */
  for (i = 0; i < height; i+=gimp_tile_height()) {
    int scanlines;
  
    /* read the next row of tiles */
    scanlines = MIN(gimp_tile_height(), height - i);
    gimp_pixel_rgn_get_rect(&pixel_rgn, in_buf, 0, i, width, scanlines);
    in_buf_pos = in_buf;
      
    /* process each pixel row */
    for (j=0; j<scanlines; j++) {
      
      /* do each pixel in the row */
      for (k=0; k<width; k++) {
	pixel_save (out_buf_pos, in_buf_pos, &olen, &ilen, OP_NORMAL);
	in_buf_pos  += ilen;
	out_buf_pos += olen;

	if (out_buf_pos >= out_buf_threshhold) {
	  fwrite (out_buf, out_buf_pos - out_buf, 1, tgt);
	  out_buf_pos = out_buf;
	}
      }
      
      /* kick the progress bar */
      gimp_progress_update ((double) (i+j) / (double) height);
    }
  } 

  pixel_save (out_buf_pos, NULL, &olen, &ilen, OP_DONE);
  out_buf_pos += olen;
  
  if (out_buf_pos > out_buf) {
    fwrite (out_buf, out_buf_pos - out_buf, 1, tgt);
  }

  
  rc = TRUE;

cleanup:
  if (out_buf) g_free (out_buf);
  if (in_buf) g_free (in_buf);

  return rc;
}


/**********************************************************/
static gint
drawable_load_lrud (FILE *src, PixelFn pixel_load, GDrawable *drawable)
{
  GPixelRgn      pixel_rgn;

  guchar         *out_buf = NULL;
  guchar         *out_buf_pos;
  guchar         *in_buf = NULL;
  guchar         *in_buf_pos;
  guchar         *in_buf_threshhold;
  gint           in_buf_size;                     /* bude gint vzdy stacit?? */

  gboolean       rc = FALSE;
  gint           i, j, k, olen, ilen;


  in_buf_size = gimp_tile_height()*width*4;             /* g()*MAX(height,width)*4 ? */
  if ((in_buf = g_new(guchar, in_buf_size)) == NULL)
    goto cleanup;

  in_buf_pos = in_buf;
  in_buf_threshhold = &in_buf[in_buf_size-5];

  i = gimp_tile_height()*width*bpp;           /* g()*MAX(height,width)*bpp ? */
  if ((out_buf = g_new(guchar, i)) == NULL)
    goto cleanup;
  out_buf_pos = out_buf;

  /* allocate a pixel region to work with */
  gimp_pixel_rgn_init (&pixel_rgn, drawable,
                       0, 0,
                       width, height,
                       TRUE, FALSE);


  fread (in_buf, in_buf_size, 1, src);                  /* omezit podle skutecne velikosti!! */

  pixel_load (NULL, in_buf_pos, &olen, &ilen, OP_INIT);
  in_buf_pos += ilen;


  /* process each row of tiles */
  for (i = 0; i < height; i+=gimp_tile_height()) {
    int scanlines = MIN(gimp_tile_height(), height - i);
      
    /* process each pixel row */
    for (j=0; j<scanlines; j++) {
      
      /* do each pixel in the row */
      for (k=0; k<width; k++) {
	if (in_buf_pos >= in_buf_threshhold) {
	  ilen = in_buf + in_buf_size - in_buf_pos;
	  memcpy(in_buf, in_buf_pos, ilen);
	  fread (in_buf+ilen, in_buf_size-ilen, 1, src);
	  in_buf_pos = in_buf;
	}

	pixel_load (out_buf_pos, in_buf_pos, &olen, &ilen, OP_NORMAL);
	out_buf_pos += olen;
	in_buf_pos  += ilen;
      }
      
      /* kick the progress bar */
      gimp_progress_update ((double) (i+j) / (double) height);
    }

    /* write the next row of tiles */
    gimp_pixel_rgn_set_rect(&pixel_rgn, out_buf, 0, i, width, scanlines);
    out_buf_pos = out_buf;
  } 
  
  pixel_load (out_buf_pos, NULL, &olen, &ilen, OP_DONE);
  out_buf_pos += olen;

#if 0
  /* tohle asi nebude fungovat!!! */
  if (out_buf_pos >  out_buf)
    gimp_pixel_rgn_set_rect(&pixel_rgn, out_buf, 0, i, width, scanlines);
#endif


  rc = TRUE;

cleanup:
  if (in_buf) g_free (in_buf);
  if (out_buf) g_free (out_buf);

  return rc;
}

/**********************************************************/
static gint
image_save_bt (FILE *tgt, DrawableFn drawable_save, PixelFn pixel_save, gint32 image_ID, GDrawable *drawable)
{
  return FALSE;
}

/**********************************************************/
static gint
image_save_tb (FILE *tgt, DrawableFn drawable_save, PixelFn pixel_save, gint32 image_ID, GDrawable *drawable)
{
  return FALSE;
}

/**********************************************************/
static gint
image_save_single (FILE *tgt, DrawableFn drawable_save, PixelFn pixel_save, gint32 image_ID, GDrawable *drawable)
{
  gint   rc;

  rc = drawable_save (tgt, pixel_save, drawable);
  return rc;
}


/**********************************************************/
static gint
image_load_bt (FILE *src, DrawableFn drawable_load, PixelFn pixel_load, gint32 image_ID, GDrawable *drawable)
{
  return FALSE;
}

/**********************************************************/
static gint
image_load_tb (FILE *src, DrawableFn drawable_load, PixelFn pixel_load, gint32 image_ID, GDrawable *drawable)
{
  return FALSE;
}

/**********************************************************/
static gint
image_load_single (FILE *src, DrawableFn drawable_load, PixelFn pixel_load, gint32 image_ID, GDrawable *drawable)
{
  gint32    layer;
  gint      rc;

  /*
  if (!variable_get_int ("layer_type", &layer_type)) layer_type = RGB_IMAGE;
  */

  layer = gimp_layer_new (image_ID, "layer", width, height, 2*image_type + alpha, 100.0, NORMAL_MODE);
  gimp_image_add_layer (image_ID, layer, -1);
  drawable = gimp_drawable_get (layer);

  /* output */
  rc = drawable_load (src, pixel_load, drawable);
  gimp_drawable_detach (drawable);

  return rc;
}

/**********************************************************/
/* End of file data.c */
