/* The GIMP -- an image manipulation program
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * RAW plugin version 0.0.1
 * Copyright (C) 1999 Jaroslav Benkovsky, <benkovsk@pha.pvt.cz>
 *
 * RCS: $Header: /home/benkovsk/worx/RAW/RCS/expression.c,v 1.3 1999/01/31 22:33:05 benkovsk Exp benkovsk $
 */

/**********************************************************/
#include <ctype.h>
#include <errno.h>

#include "RAW.h"

/**********************************************************/
gchar*
string_escape (gchar *text, gint len)
{
  gchar   buf[MAX_INBUF_SIZE];
  gchar   *p;
  guchar  c;

  p = buf;
  for ( ; len; len--) {
    c = *text++;
    if (c < ' ' || c > 127 || c == '"') {
      *p++ = '\\';

      switch (c) {
      case '\n': c = 'n'; break;
      case '\t': c = 't'; break;
      case '\r': c = 'r'; break;
      case '"' : c = '"'; break;
      default: 
	*p++ = (c / 64) + '0';
	c %= 64;
	*p++ = (c / 8) + '0';
	c = (c % 8) + '0' ;
	break;
      }
    }
    *p++ = c;
  }
  *p++ = '\0';

  return g_strdup (buf);
}

/**********************************************************/
gchar*
string_unescape (gchar *text, gint *len)
{
  gchar   buf[MAX_INBUF_SIZE];
  gchar   *p;
  guchar  c;
  gint    l;
  gint    i, a;

  p = buf;

  for (c = *text; c; c = *++text) {
    if (c=='\\') {
      c = *++text;
      switch (c) {
      case 'n': c = '\n'; break;
      case 't': c = '\t'; break;
      case 'r': c = '\r'; break;
      case 'x':
	c = 0;
	for (i=0; i<2; i++) {
	  a = toupper (*++text);
	  if (a >= '0' && a <= '9') a -= '0';
	  else if (a >= 'A' && a <= 'F') a = a - 'A' + 10;
	  else { text--; break; }

	  c = c*16 + a;
	}
	break;
	
      default:
	if (c >='0' && c <= '7') {
	  c -= '0';
	  for (i=0; i<2; i++) {
	    a = *++text;
	    if (a >= '0' && a <= '7') a -= '0';
	    else { text--; break; }

	    c = c*8 + a;
	  }
	  break;
	}
	
      }  /* switch (c) */
    }  /* c == '\\' */
    *p++ = c;
  }

  l = p - buf;
  text = (VarText*) g_malloc (l + 1);
  memcpy (text, buf, l);
  *(text + l) = 0;
  *len = l;

  return text;
}

/**********************************************************/
/* Free whole expression from memory */
/**********************************************************/
void
expression_free (Expression *value)
{
  if (!value) return;

  g_free (value->text);
  g_free (value);
}

/**********************************************************/
Expression*
expression_from_string (gchar *text, gchar **newtext)
{
  char       c;
  char       *buf_p;
  char       *expr = text;
  int        len, depth;
  Expression  *value;

  value = g_new0 (Expression, 1);
  c = toupper (*expr);

  /************** EXP_NUM or EXP_FLOAT **************/
  if (isdigit (c) || c == '+' || c == '-') {
    len = strspn (expr + 1, "0123456789");
    c = *(expr + len + 1);
    errno = 0;
    if (c == '.' || toupper (c) == 'E') {
      value->type = EXP_FLOAT;
      value->fnum = strtod (expr, &expr);
    }
    else {
       value->type = EXP_NUM;
       value->num = strtol (expr, &expr, 0);
    }
    if (errno) {
      WARNING ("Invalid number: %s\n", expr);
      expr += len;
    }
  }
  /************** EXP_STR **************/
  else if (c == '"') {
    value->type = EXP_STR;

    while (0 != (c = *++expr)) {
      if (c == '\\' && *(expr + 1)) ++expr;
      else if (c == '"') break;
    }

    if (!c) WARNING ("Unterminated string: %s\n", text);
    else    *expr = 0;

    value->text = string_unescape (text + 1, &value->len);
    *expr = c;
    expr++;
  }
  /************** EXP_VAR **************/
  else if (c >= 'A' && c <= 'Z') {
    value->type = EXP_VAR;

    len = strspn (expr, "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_-?");
    value->text = g_malloc (len + 1);
    memcpy(value->text, expr, len);
    *(value->text+len) = 0;
    value->len = len;
    expr += len;
  }
  /************** EXP_LISP **************/
  else if (c == '(') {
    value->type = EXP_LISP;
    depth=1;
    buf_p = expr;
    while (*buf_p && depth) {
      c=*++buf_p;
      if (c=='(') depth++;
      else if (c==')') depth--;
      /* FIXME: parentheses in strings etc. */
    }
    len = buf_p-expr+1;

    value->text = g_malloc (len + 1);
    memcpy(value->text, expr, len);
    *(value->text + len) = 0;
    value->len = len;
    expr = buf_p;
  }
  /************** error **************/
  else {
    ERROR ("Unrecognized expression: %s\n", expr);
  }

  if (newtext) *newtext = expr;
  return value;
}

/**********************************************************/
gboolean
expression_assign_from_string (Expression *value, gchar *text)
{
  switch (value->type) {
  case EXP_NUM: 
  case EXP_FLOAT:
    expression_assign_value_from_string (value, text);
    break;

  case EXP_STR:
    /* FIXME: what about '"' ? */
    if (value->text) g_free (value->text);
    value->text = string_unescape (text, &value->len);
    break;

  case EXP_LISP:
  case EXP_VAR:
    if (value->text) g_free (value->text);
    value->text = g_strdup (text);
    value->len = strlen (text);
    break;

  default:
    g_assert_not_reached ();
    return FALSE;
  }

  return TRUE;
}

/**********************************************************/
gboolean
expression_assign_value_from_string (Expression *value, gchar *text)
{
  switch (value->type) {
  case EXP_NUM: 
    value->num = strtol (text, NULL, 0);
    break;

  case EXP_FLOAT:
    value->fnum = strtod (text, NULL);
    break;

  case EXP_STR:
    if (value->text) g_free (value->text);
    value->text = string_unescape (text, &value->len);
    break;

  default:
    g_assert_not_reached ();
    return FALSE;
  }

  /* FIXME: should check errno for strto* */
  return TRUE;
}

/**********************************************************/
/* dumps value into string which can be read lately */
/**********************************************************/
gchar*
expression_to_string (Expression *value)
{
  gchar   *res;
  gchar   *text;

  if (!value) return NULL;

  switch (value->type) {
  case EXP_NUM: 
  case EXP_FLOAT:
    res = expression_value_to_string (value);
    break;

  case EXP_STR: 
    text = string_escape (value->text, value->len);
    res = g_malloc (strlen (text) + 3);
    res[0]='"';
    strcpy (res+1, text);
    strcat (res, "\"");
    g_free (text);
    break;

  case EXP_VAR:
  case EXP_LISP:
    res = g_strdup (value->text);
    break;

  default:
    g_assert_not_reached ();
    return NULL;
  }

  return res;
}

/**********************************************************/
gchar*
expression_value_to_string (Expression *value)
{
  gchar   buf[30];
  gchar   *res;

  if (!value) return NULL;

  switch (value->type) {
  case EXP_STR: 
    res = string_escape (value->text, value->len);
    break;

  case EXP_NUM: 
    sprintf (buf, "%d", value->num); 
    res = g_strdup (buf);
    break;

  case EXP_FLOAT:
    sprintf (buf, "%f", value->fnum);
    res = g_strdup (buf);
    break;

  default:
    g_assert_not_reached ();
    return NULL;
  }

  return res;
}


/**********************************************************/
ExpressionType
expression_evaluate (Expression *value, VarNum *num, VarText **text, gint *len)
{
  Expression      *expr = value;
  ExpressionType  type;
  gchar           *varname;

#ifdef RAW_WITH_GUILE
  SCM             scm;
#endif


  if (value->type == EXP_VAR) {
    varname = (gchar*)value->text;

    expr = variable_get (varname);
    if (!expr) {
      WARNING ("Variable '%s' not found\n", varname);
      *num = 0;
      return EXP_NUM;
    }
  }

  switch (expr->type) {
  case EXP_NUM:  *num = expr->num; 
                 type = EXP_NUM;
		 break;

#if 0      /* FIXME */
  case EXP_FLOAT: *fnum = expr->fnum; 
                  type = EXP_FLOAT;
		  break;
#endif

  case EXP_STR:  *text = (gchar*) g_malloc (expr->len+1); 
                 memcpy (*text, expr->text, expr->len+1);
		 *len = expr->len;
		 type = EXP_STR;
		 break;

#ifdef RAW_WITH_GUILE
  case EXP_LISP: scm = gh_eval_str ((gchar*)expr->text);

                 if (gh_exact_p (scm)) {
		   *num = SCM2NUM (scm);
		   type = EXP_NUM; 
		 }  /* FIXME: EXP_FLOAT */
		 else {
		   *text = gh_scm2newstr (scm, len);
		   type = EXP_STR;
		 }
		 break;
#endif

  default:       g_assert_not_reached ();
                 return 0;       /* FIXME: prevents compiler warning */
  }

  if (value != expr) expression_free (expr);
  return type;
}

/**********************************************************/
/* if lvalue is variable, assigns the variable either text or
   num (if text == NULL). If lvalue
   is anything else, compares it to num/text and return TRUE
   if match, FALSE otherwise. It serves for validating
   file header contents. */
/**********************************************************/

gboolean
expression_assign (Expression *lvalue, VarNum num, VarText *text, gint len)
{
  switch (lvalue->type) {
  case EXP_NUM:  return (num == lvalue->num);

#if 0   /* FIXME */
  case EXP_FLOAT: return (fnum == lvalue->fnum);
#endif

  case EXP_STR:  if (!text || !lvalue->text) return FALSE;
                 if (len != lvalue->len) return FALSE;
                 return (0 == memcmp (text, lvalue->text, len));

  case EXP_VAR:  if (text) variable_text_add (lvalue->text, text, len);  
                 else variable_num_add (lvalue->text, num);
                 /* FIXME: check if exists? */
                 /* FIXME: EXP_FLOAT */
                 return TRUE;

#ifdef RAW_WITH_GUILE
  case EXP_LISP: 
#if 0            /* FIXME: maybe it should assign, not compare */
                 {
		   SCM    scm = gh_eval_str ((gchar*)value->text);

		   if (gh_exact_p (scm)) {
		     return (num == SCM2NUM (scm));
		   }
		   /* FIXME: EXP_FLOAT */
		   else {
		     gint  l;
		     VarText  *t = gh_scm2newstr (scm, l);

		     if (!text || !t) return FALSE;
		     if (len != l) return FALSE;
		     res = memcmp (text, t, len);
		     g_free (t);
		     return !res;
		   }
		 }
#endif    /* 0 */
             return TRUE;
#endif    /* RAW_WITH_GUILE */

  default:   g_assert_not_reached ();
             return FALSE;              /* FIXME: prevents compiler warning */
  }
}

/**********************************************************/
Expression*
expression_copy (Expression *expr)
{
  Expression  *value;


  if (!expr) return NULL;

  value = g_new0 (Expression, 1);
  memcpy (value, expr, sizeof (Expression));

  switch (expr->type) {
  case EXP_NUM  :
  case EXP_FLOAT: break;

  case EXP_STR  :
  case EXP_LISP :
  case EXP_VAR  : value->text = g_malloc (expr->len + 1);
                  memcpy (value->text, expr->text, expr->len);
		  value->text[expr->len] = 0;
		  break;

  default       : g_assert_not_reached ();
  }

  return value;
}

/**********************************************************/
gint
expression_type_to_pdb (ExpressionType type)
{
  switch (type) {
  case EXP_NUM: 
    return PARAM_INT32;

  case EXP_FLOAT: 
    return PARAM_FLOAT;

  case EXP_VAR:
  case EXP_LISP:
  case EXP_STR: 
    return PARAM_STRING;

  default: 
    g_assert_not_reached ();
    return 0;
  }
}
/**********************************************************/
#ifdef DEBUG

void
expression_print (char *name, Expression *value)
{
  char   *s = "???";
  char   *expr;

  switch (value->type) {
  case EXP_NUM  : s = "num"; break;
  case EXP_FLOAT: s = "float"; break;
  case EXP_STR  : s = "str"; break;
  case EXP_VAR  : s = "var"; break;
  case EXP_LISP : s = "lisp"; break;
  }

  g_print ("Exp: %s [%s] = ", name, s);
  expr = expression_to_string (value);
  g_print ("%s\n", expr);
  g_free (expr);
}

#endif  /* DEBUG */


/**********************************************************/
/* End of file expression.c */
