/* The GIMP -- an image manipulation program
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * RAW plugin version 0.0.1 
 * Copyright (C) 1999 Jaroslav Benkovsky, <benkovsk@pha.pvt.cz>
 *
 * RCS: $Header: /home/benkovsk/worx/RAW/RCS/subplugin.c,v 1.2 1999/01/31 22:31:53 benkovsk Exp benkovsk $
 */

/**********************************************************/
#include <string.h>

#include "RAW.h"

/**********************************************************/
static gboolean
subplugin_install (RawFormat *format);

static void
subplugin_build_args (RawFormat *format, ContextType context, gboolean ret, 
		      GParamDef **args, gint *nargs);

/**********************************************************/
void
subplugin_install_all (void)
{
  GList  *flist = rawformat_list->next;

  while (flist) {
    RawFormat  *format = flist->data;
    flist = flist->next;

    subplugin_install (format);
  }
}

/**********************************************************/
gchar*
subplugin_get_name (RawFormat *format, ContextType context)
{
  gchar  *name;
  gchar  *plugin_name;

  g_assert ((context == CTX_LOAD) || (context == CTX_SAVE));

  if (context == CTX_SAVE) {
    plugin_name = PLUGIN_SAVE_NAME;
  }
  else {
    plugin_name = PLUGIN_LOAD_NAME;
  }

  name = g_malloc (strlen (plugin_name) + strlen (format->id) + 2);
  strcpy (name, plugin_name);
  strcat (name, "_");
  strcat (name, format->id);

  return name;
}

/**********************************************************/
/* load subplugin args saved via PDB
 */

gboolean
subplugin_get_args_from_pdb (RawFormat *format, ContextType context)
{
  gchar     *sp_name = subplugin_get_name (format, context);
  gchar     *sp_args = (gchar*)g_malloc (format->max_arg_size+1);
  gboolean  rc = TRUE;

  sp_args[0] = 0;
  gimp_get_data (sp_name, sp_args);

  if (sp_args[0])
    rc = subplugin_get_args_from_string (format, context, sp_args);

  g_free (sp_name);
  g_free (sp_args);
  return rc;
}

/**********************************************************/
/* save subplugin args saved via PDB
 */

void
subplugin_set_args_to_pdb (RawFormat *format, ContextType context)
{
  gchar  *sp_name = subplugin_get_name (format, context);
  gchar  *sp_args = subplugin_set_args_to_string (format, context);

  gimp_set_data (sp_name, sp_args, strlen (sp_args) + 1);
  g_free (sp_name);
  g_free (sp_args);
}

/**********************************************************/
static void
subplugin_build_args (RawFormat *format, ContextType context, gboolean ret, 
		      GParamDef **args, gint *nargs)
{
  static GParamDef load_args[] =
  {
    { PARAM_INT32, "run_mode", "Interactive, non-interactive" },
    { PARAM_STRING, "filename", "The name of the file to load" },
    { PARAM_STRING, "raw_filename", "The name entered" },
  };
  static GParamDef load_return_vals[] =
  {
    { PARAM_IMAGE, "image", "Output image" },
  };
  static int nload_args = sizeof (load_args) / sizeof (load_args[0]);
  static int nload_return_vals = sizeof (load_return_vals) / sizeof (load_return_vals[0]);

  static GParamDef save_args[] =
  {
    { PARAM_INT32,    "run_mode", "Interactive, non-interactive" },
    { PARAM_IMAGE,    "image", "Input image" },
    { PARAM_DRAWABLE, "drawable", "Drawable to save" },
    { PARAM_STRING,   "filename", "The name of the file to save the image in" },
    { PARAM_STRING,   "raw_filename", "The name of the file to save the image in" },
  };
  static int nsave_args = sizeof (save_args) / sizeof (save_args[0]);

  GParamDef  *iargs, *def_args;
  gint       ndef_args, nask_args, i;
  GList      *alist;

  g_assert ((context == CTX_SAVE) || (context == CTX_LOAD));

  iargs = NULL;
  def_args = NULL;
  ndef_args = 0;
  nask_args = 0;

  if (ret) {
    if (context == CTX_LOAD) {
      ndef_args = nload_return_vals;
      def_args = load_return_vals;
    }
  }
  else {
    if (context == CTX_SAVE) {
      def_args  = save_args;
      ndef_args = nsave_args;
      nask_args = format->ask_cnt_save;
    }
    else {
      def_args  = load_args;
      ndef_args = nload_args;
      nask_args = format->ask_cnt_load;
    }
  }

  if (ndef_args + nask_args) {
    iargs = g_new0 (GParamDef, (ndef_args + nask_args));

    for (i=0; i<ndef_args; i++) {
      iargs[i].type = def_args[i].type;
      iargs[i].name = def_args[i].name;
      iargs[i].description = def_args[i].description;
    }

    if (ret) alist = NULL;
    else     alist = format->ask_list->next;

    while (alist) {
      AskItem  *ask = alist->data;

      alist = alist->next;

      if (ask->context == context || ask->context == CTX_BOTH) {
	g_assert (i<(ndef_args + nask_args)); /* FIXME */
	iargs[i].type = expression_type_to_pdb (ask->value->type);
	iargs[i].name = ask->name;
	iargs[i].description = ask->name;
	i++;
      }
    }
  }

  *args = iargs;
  *nargs = ndef_args + nask_args;
}

/**********************************************************/
static gboolean
subplugin_install (RawFormat *format)
{
  gchar        *name;
  gchar        *blurb;
  gchar        *desc;
  gchar        *author;
  gchar        *maintainer;
  gchar        *date;
  gchar        *menu;
  gchar        *ext;
  gchar        *magic;
  gchar        *typestr;
  const gchar  *blurb_fmt = "Loads or saves image in raw file format \"%s\"";

  GParamDef    *args;
  GParamDef    *rargs;
  gint         nargs, nrargs;

  if (!format->id) return FALSE;

  /* save & load */
  blurb = (gchar*)g_malloc (strlen (blurb_fmt) + strlen (format->name));
  sprintf (blurb, blurb_fmt, format->name);

  desc = format->desc ? format->desc : "No help defined";
  author = format->author ? format->author : "(RAW plugin: Jaroslav Benkovsky)";
  maintainer = format->maintainer ? format->maintainer : author;
  date = format->date ? format->date : "";

  ext = format->ext ? format->ext : NULL;
  magic = format->magic ? format->magic : NULL;

  typestr = format->type ? format->type : "*";


  /* save */
  name = subplugin_get_name (format, CTX_SAVE);
  menu = format->savemenu ? format->savemenu : NULL;

  subplugin_build_args (format, CTX_SAVE, FALSE, &args, &nargs);
  subplugin_build_args (format, CTX_SAVE, TRUE, &rargs, &nrargs);

  gimp_install_procedure (name,
                          blurb,
                          desc,
                          author,
                          maintainer,
                          date,
                          menu,
                          typestr,
                          PROC_PLUG_IN,
			  nargs, nrargs,
                          args, rargs);

  if (ext && menu) gimp_register_save_handler (name, ext, menu);

  if (args) g_free (args);
  if (rargs) g_free (rargs);
  g_free (name);


  /* load */
  name = subplugin_get_name (format, CTX_LOAD);
  menu = format->loadmenu ? format->loadmenu : NULL;

  subplugin_build_args (format, CTX_LOAD, FALSE, &args, &nargs);
  subplugin_build_args (format, CTX_LOAD, TRUE, &rargs, &nrargs);

  gimp_install_procedure (name,
                          blurb,
                          desc,
                          author,
                          maintainer,
                          date,
                          menu,
                          NULL,
                          PROC_PLUG_IN,
			  nargs, nrargs,
                          args, rargs);

  if (ext && menu && magic) gimp_register_magic_load_handler (name, ext, menu, magic);

  if (args) g_free (args);
  if (rargs) g_free (rargs);
  g_free (name);


  g_free (blurb);

  return TRUE;
}

/**********************************************************/
/* function to get format subplugin's arguments when run in
 * in noninteractive mode
 */

gboolean
subplugin_get_args (RawFormat *format, 
		    ContextType context, 
		    GParam *args, 
		    gint nargs)
{
  GList   *alist;
  gint    i, cnt;

  g_assert ((context == CTX_LOAD) || (context == CTX_SAVE));

  if (context == CTX_SAVE) {
    cnt = format->ask_cnt_save;
  }
  else {
    cnt = format->ask_cnt_load;
  }

  if (nargs != cnt) {
    WARNING ("Expected %d subplugin args, got %d\n", cnt, nargs);
    return FALSE;
  }

  alist = format->ask_list->next;
  i = 0;
  while (alist) {
    AskItem   *ask = (AskItem*) alist->data;
    alist = alist->next;

    if (ask->context == context || ask->context == CTX_BOTH) {
      g_assert (i<nargs);               /* FIXME: should be warning */


      if (expression_type_to_pdb (ask->value->type) != args[i].type) {
	WARNING ("Wrong type of PDB argument\n");
	return FALSE;
      }



      /* !!!!!!!!!!!!!!!!!!!!!!!!!!! */

      if (ask->value->type == EXP_STR) { /* FIXME EXP_FLOAT */
	if (ask->value->text) g_free (ask->value->text);
	ask->value->text = string_unescape (args[i].data.d_string, &ask->value->len);
	
	variable_text_add (ask->name, ask->value->text, ask->value->len);
	
      }
      else {
	ask->value->num = args[i].data.d_int32;
	
	variable_num_add (ask->name, args[i].data.d_int32);
	
      }
      i++;
    }
  }

  return TRUE;
}

/**********************************************************/
gboolean
subplugin_get_args_from_string (RawFormat *format, 
				ContextType context, 
				gchar *argstr)
{
  GParam      *args;
  Expression  *expr;
  gint        i, cnt;
  gboolean    rc;


  g_assert ((context == CTX_LOAD) || (context == CTX_SAVE));

  if (context == CTX_SAVE) {
    cnt = format->ask_cnt_save;
  }
  else {
    cnt = format->ask_cnt_load;
  }

  args = g_new0 (GParam, cnt);


  /* FIXME: it should probably use Expressions directly and 
   * do not xlat to PDB-style args 
   */

  for (i=0; argstr[0] != '\0' && i<cnt; i++) {
    expr = expression_from_string (argstr, &argstr);
    if (argstr[0] == ':') argstr++;

    if (expr) {
      switch (expr->type) {
      case EXP_STR : 
	args[i].type = PARAM_STRING;
	args[i].data.d_string = expr->text;
	expr->text = NULL;
	break;

      case EXP_NUM:
	args[i].type = PARAM_INT32;
	args[i].data.d_int32 = expr->num;
	break;

      default:
	ERROR ("Invalid argument");
	return FALSE;
      }
      expression_free (expr);
    }
    else ERROR ("Can't read argument");
  }

  if (i!=cnt || argstr[0]!='\0') ERROR ("Wrong number of arguments");

  rc = TRUE;
  if (cnt) 
    rc = subplugin_get_args (format, context, args, cnt);

  for (i=0; i<cnt; i++) {
    if (args[i].type == PARAM_STRING) g_free (args[i].data.d_string);
  }
  g_free (args);

  return rc;
}

/**********************************************************/
gchar*
subplugin_set_args_to_string (RawFormat *format, 
			      ContextType context)
{
  GList        *alist;
  Expression   *expr;
  gint         i, cnt, total_len;
  gchar        **args;
  gchar        *res;

  g_assert ((context == CTX_LOAD) || (context == CTX_SAVE));

  if (context == CTX_SAVE) {
    cnt = format->ask_cnt_save;
  }
  else {
    cnt = format->ask_cnt_load;
  }

  args = g_new (gchar*, cnt);

  alist = format->ask_list->next;
  i = 0;
  total_len = 0;

  while (alist) {
    AskItem   *ask = (AskItem*) alist->data;
    alist = alist->next;

    if (ask->context == context || ask->context == CTX_BOTH) {
      g_assert (i<cnt);

      expr = variable_get (ask->name);

      if (expr) {
	if (ask->value->type != expr->type)
	  WARNING ("Wrong arg type: %d != %d", 
		   ask->value->type, expr->type); /* FIXME: error? */

	args[i] = expression_to_string (expr);
	total_len += strlen (args[i]);
      }
      i++;
    }
  }

  g_assert (i==cnt);

  res = g_malloc (total_len + cnt + 1);
  res[0] = '\0';
  for (i=0; i<cnt; i++) {
    if (i) strcat (res, ":");
    strcat (res, args[i]);
    g_free (args[i]);
  }

  return res ;
}

/**********************************************************/
void
subplugin_set_vars (RawFormat *format, ContextType context)
{
  GList  *alist = format->ask_list->next;
  while (alist) {
    AskItem  *ask = (AskItem*)alist->data;
    alist = alist->next;

    if (ask->context == context || ask->context == CTX_BOTH)
      variable_expr_add (ask->name, ask->value);
  }
}

/**********************************************************/
/* End of file subplugin.c */
